/*
 * (C) Copyright IBM Corporation 2016.
 * (C) Copyright HCL Technologies Limited 2016, 2017.
*/
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA

 Author: IBM Corporation and/or HCL Technologies Ltd.
 This module is part of the IBM (R) Rational (R) ClearCase (R)(*)
 Multi-version file system (MVFS).
 For support, please visit http://www.ibm.com/software/support

 * ClearCase is a trademark of IBM Corporation, registered in many
   jurisdictions, and is used under license.
*/
#include "mvfs_systm.h"
#include "mvfs.h"

/* These entries provide the machine dependent routines for the corresponding
** MVFS_LPC_MDEP_* macros defined in mvfs_lpc_linux.h (which is included by
** mvfs_lpc_mdep.h included in mvfs_lpc.c) and used by the "generic" mvfs_lpc_*
** routines in mvfs_lpc.c
**
** On Linux, we're going to use an "inverted call" through the
** MVFS_CMD_LPC_MESSAGE IOCTL to transport the RPC data between the kernel and
** user space.  User space creates a buffer and calls the IOCTL, which then
** waits in the kernel until MVFS needs to send an RPC.  The MVFS fills in the
** buffer with the RPC request and causes the IOCTL to return.  User space
** processes the RPC request and generates a reply in a buffer which it then
** calls the IOCTL with.  Back in the kernel, the reply is returned to MVFS for
** the RPC request it made and the IOCTL waits (with its buffer) for the next
** RPC request.
*/

/* Requires the caller to hold the handle lock (h->ic_h_lock). */
static void
mvfs_lpc_list_add(
    MVFS_LPC_MDEP_H_T h,
    ic_krn_req_t *add_p)
{
    ASSERT((h != NULL) && (add_p != NULL));

    switch (add_p->ickr_type) {
      case IC_PENDING:          /* FIFO list. */
        if (h->ic_pend_tail == NULL) {
            ASSERT(h->ic_pend_head == NULL); /* Empty list. */
            h->ic_pend_head = add_p;
            h->ic_pend_tail = add_p;
        } else {
            (h->ic_pend_tail)->ickr_next = add_p;
            h->ic_pend_tail = add_p;
        }
        MVFS_LPC_LOG("Pending op=%d mh=%lld (head=%p tail=%p)\n",
                     add_p->ickr_op, add_p->ickr_handle,
                     h->ic_pend_head, h->ic_pend_tail);
        break;

      case IC_SENT:             /* LIFO list. */
        add_p->ickr_next = h->ic_sent_head;
        h->ic_sent_head = add_p;
        MVFS_LPC_LOG("Sent op=%d mh=%lld (head=%p)\n",
                     add_p->ickr_op, add_p->ickr_handle, h->ic_sent_head);
        break;

      default:
        mvfs_log(MFS_LOG_ERR,
                 "%s: Entry to add=%p to handle=%p has bad type=%d\n",
                 __func__, add_p, h, add_p->ickr_type);
        break;
    }
}

/* Requires the caller to hold the handle lock (h->ic_h_lock).  Remove the
** specified request from the list that it's on (given by the ickr_type field).
*/
static void
mvfs_lpc_list_remove_req(
    MVFS_LPC_MDEP_H_T h,
    ic_krn_req_t *remove_p)
{
    ic_krn_req_t **link_pp;
    ic_krn_req_t *prev_p;

    ASSERT(h != NULL);

    /* Setup to remove the request from whichever list it is on. */
    if (remove_p->ickr_type == IC_PENDING) {
        link_pp = &(h->ic_pend_head);
        prev_p = NULL;
    } else if (remove_p->ickr_type == IC_SENT) {
        link_pp = &(h->ic_sent_head);
    } else {
        mvfs_log(MFS_LOG_ERR,
                 "%s: Entry to remove=%p from h=%p has bad type=%d,"
                 " continuing...\n",
                 __func__,
                 remove_p,
                 h,
                 remove_p->ickr_type);
        goto done;
    }
    /* Find the request on the specified list.  We keep the prev_p, even though
    ** we only need it for the pending list (because it is managed as a FIFO
    ** with a tail pointer), to simplify the code.
    */
    while ((*link_pp != NULL) && (*link_pp != remove_p)) {
        prev_p = *link_pp;
        link_pp = &((*link_pp)->ickr_next);
    }
    if (*link_pp == NULL) {
        mvfs_log(MFS_LOG_INFO,
                 "%s: LPC op=%d (%p) should be on the %s list,"
                 " h=%p, continuing....\n",
                 __func__,
                 remove_p->ickr_op,
                 remove_p,
                 (remove_p->ickr_type == IC_SENT) ? "sent" : "pending",
                 h);
        goto done;
    }
    /* The entry was found, remove it. */
    *link_pp = remove_p->ickr_next;
    remove_p->ickr_next = NULL;

    /* Update the pending list tail ptr if we remove what it points at. */
    if (remove_p->ickr_type == IC_PENDING) {
        if (h->ic_pend_tail == remove_p) {
            h->ic_pend_tail = prev_p;
        }
    }
    MVFS_LPC_LOG("%s op=%d mh=%lld (head=%p tail=%p)\n",
                 (remove_p->ickr_type == IC_SENT) ? "Sent" : "Pending",
                 remove_p->ickr_op,
                 remove_p->ickr_handle,
                 h->ic_pend_head,
                 h->ic_pend_tail);
  done:
    return;
}

/* Requires the caller to hold the handle lock (h->ic_h_lock).  Remove the
** request with the specified handle from the sent list and return the request
** we found, or NULL if we didn't find one that matched.
*/
static ic_krn_req_t *
mvfs_lpc_list_remove_id(
    MVFS_LPC_MDEP_H_T h,
    uint64_t user_handle)
{
    ic_krn_req_t **link_pp;
    ic_krn_req_t *kreqp;

    ASSERT(h != NULL);

    /* Look for a request with the user_handle on the sent list. */
    link_pp = &(h->ic_sent_head);

    while (*link_pp != NULL && (*link_pp)->ickr_handle != user_handle) {
        link_pp = &((*link_pp)->ickr_next);
    }
    if ((kreqp = *link_pp) != NULL) {
        /* Found it, remove it from the list. */
        *link_pp = kreqp->ickr_next;
        kreqp->ickr_next = NULL;
        MVFS_LPC_LOG("Remove sent op=%d mh=%lld (head=%p)\n",
                     kreqp->ickr_op, user_handle, h->ic_sent_head);
    }
    return kreqp;               /* Could be NULL. */
}

/* Forward decls for the routines we use below. */

static void
mvfs_lpc_linux_match_res_to_req(
    MVFS_LPC_MDEP_H_T h,
    mvfs_lpc_usr_hdr_t *uhdrp,
    mvfs_lpc_usr_msg_t *umsgp,
    u_long datalen,
    CALL_DATA_T *cd
);

static int
mvfs_lpc_linux_wait_for_req(
    MVFS_LPC_MDEP_H_T h,
    mvfs_lpc_usr_hdr_t *uhdrp,
    mvfs_lpc_usr_msg_t *umsgp,
    u_long datalen,
    CALL_DATA_T *cd
);

/* Called from mvfs_vmount_subr at automatic viewroot mount time to do any
** global initialization needed by the LPC on this platform.  The viewroot mount
** is just a normal file system mount.  All the necessary LPC initialization
** (for communicating with the rview_agent that manages a view) is done when the
** viewtag is created (in mvfs_lpc_linux_create below).
*/
int
mvfs_lpc_linux_initialize(void)
{
    int error = 0;

    return error;
}

/* Called from mvfs_vunmount_subr at automatic viewroot unmount time to undo the
** global initialization that was done at mount time for LPC on this platform.
** Nothing to undo at normal file system viewroot unmount time.
*/
int
mvfs_lpc_linux_teardown(void)
{
    int error = 0;

    return error;
}

/* Called from mvfs_mkviewtag_subr when an automatic view tag is created to do
** any per-view initialization needed for LPC on this platform.  In particular,
** it returns a handle that is used to identify the "connection" used by
** mvfs_lpc_call when an LPC for this view is made.
*/
int
mvfs_lpc_linux_create(
    const char *nm,
    CRED_T *cred,
    MVFS_LPC_MDEP_H_T *hpp
)
{
    int error = 0;
    MVFS_LPC_MDEP_H_T hp;       /* really ic_linux_h_t* in mvfs_lpc_linux.h. */

    /* Allocate the structure that the handle will refer to. */
    if ((hp = (MVFS_LPC_MDEP_H_T)mdki_linux_kmalloc(sizeof(*hp), KM_SLEEP))
        == NULL)
    {
        error = ENOMEM;
        goto done;
    }
    /* Initialize the handle structure.  Note, the lock name only uses 8 chars,
    ** and the name passed in (from mvfs_mkviewtag_subr calling mvfs_lpc_create)
    ** is the view UUID string, so it will be truncated (and the cast is to get
    ** rid of the const).  Note, this handle is part of a mvfs_lpc_handle_pvt_t,
    ** which holds the ref counting stuff and calls mvfs_lpc_destroy (which
    ** calls MVFS_LPC_MDEP_DESTROY, thence mvfs_lpc_linux_destroy) to clean this
    ** stuff up.
    */
    INITLOCK(&(hp->ic_h_lock), (char *)nm);
    hp->ic_shutting_down = FALSE;
    init_waitqueue_head(&(hp->ic_ioctl_waitqueue));
    hp->ic_pend_head = NULL;
    hp->ic_pend_tail = NULL;
    hp->ic_sent_head = NULL;

  done:
    *hpp = hp;                  /* Might be NULL. */
    return error;
}

/* The callback routine for mvfs_rcr_init (called by mvfs_lpc_create), which is
** called by mvfs_rcr_rele (called by mvfs_lpc_rele) when the reference count
** goes to 0.  It should undo anything platform specific related to the LPC
** "connection" for an automatic view (represented by the handle returned by
** mvfs_lpc_create).  Note, there were 2 holds on the handle, one for the view
** and one for the history mode view that goes with it.  So, we only get here
** when the view vnode (or hm view vnode) has been rele'd to 0 and the mnode is
** being destroyed by mvfs_mnclean, which calls mvfs_lpc_rele, which invokes
** this callback.  By this time there should be nothing on any handle list or
** waitqueues because any LPC caller or IOCTL caller would have had a hold on
** the view vnode and since its refcnt is now 0, they must have all returned
** (meaning any use they were making of the handle is done).
*/
int
mvfs_lpc_linux_destroy(MVFS_LPC_MDEP_H_T h)
{
    int err;

    /* Remember, MVFS_LPC_MDEP_H_T is really an ic_linux_h_t *. */
    MVFS_LPC_LOG("h=%p\n", h);

    if (h == NULL) {
        mvfs_log(MFS_LOG_ERR, "%s: h=NULL\n", __func__);
        err = EINVAL;
    } else if (ISLOCKED(&(h->ic_h_lock)) ||
               waitqueue_active(&(h->ic_ioctl_waitqueue)) ||
               (h->ic_pend_head != NULL) ||
               (h->ic_pend_tail != NULL) ||
               (h->ic_sent_head != NULL))
    {
        mvfs_log(MFS_LOG_ERR, "%s: lock=%d wqa=%d ph=%p pt=%p sh=%p\n",
                 __func__, ISLOCKED(&(h->ic_h_lock)),
                 waitqueue_active(&(h->ic_ioctl_waitqueue)),
                 h->ic_pend_head, h->ic_pend_tail, h->ic_sent_head);
        err = EINVAL;
    } else {
        /* There's nothing to cleanup from an init_waitqueue_head call, so just
        ** free the lock and then our handle storage.  The rest of the handle
        ** storage at mn_view.lpc_handle will be freed by our caller
        ** (mvfs_lpc_destroy).
        */
        FREELOCK(&(h->ic_h_lock));
        mdki_linux_kfree((void *)h, sizeof(*h));
        err = 0;
    }
    return err;
}

/* Send wakeups to anybody who is on the handle's pending or sent lists or is
** waiting on the ioctl waitqueue.  The woken up waiter will remove the request
** from whichever list it is on.  The handle itself will be cleaned up in
** mvfs_lpc_linux_destroy called at mvfs_mnclean time.
*/
int
mvfs_lpc_linux_flush(MVFS_LPC_MDEP_H_T h)
{
    ic_krn_req_t *kreqp;

    ASSERT(h != NULL);

    MVFS_LOCK(&(h->ic_h_lock));

    /* Tell everyone what's going on when they wakeup. */
    h->ic_shutting_down = TRUE;
    mvfs_log(MFS_LOG_DEBUG, "%s: h=%p\n", __func__, h);

    /* Wakeup each request on the sent list (wakee will remove). */
    for (kreqp = h->ic_sent_head; kreqp != NULL; kreqp = kreqp->ickr_next) {
        wake_up(&(kreqp->ickr_waitqueue));
        MVFS_LPC_LOG("Wakeup sent req=%p.\n", kreqp);
    }
    /* Wakeup each request on the pending list (wakee will remove). */
    for (kreqp = h->ic_pend_head; kreqp != NULL; kreqp = kreqp->ickr_next) {
        wake_up(&(kreqp->ickr_waitqueue));
        MVFS_LPC_LOG("Wakeup pending req=%p.\n", kreqp);
    }
    /* Wakeup everyone waiting in the ioctl waitqueue. */
    wake_up_interruptible_all(&(h->ic_ioctl_waitqueue));
    MVFS_LPC_LOG("Wakeup all ioctl h=%p wq=%p n=%p p=%p.\n",
                 h, &(h->ic_ioctl_waitqueue),
                 h->ic_ioctl_waitqueue.task_list.next,
                 h->ic_ioctl_waitqueue.task_list.prev);

    MVFS_UNLOCK(&(h->ic_h_lock));
    return 0;
}

/* Called by mvfs_vwcall to make an LPC (RPC) over the specified "connection" to
** an automatic view server (rview_agent).  This routine sends the request and
** waits for the response.  The returned status is for this call, not the
** results of the RPC (which is in the RPC response).
*/
enum clnt_stat
mvfs_lpc_linux_call(
    MVFS_LPC_MDEP_H_T h,
    int op,
    XID_T xid,
    u_long timeout,
    char *input_buf,
    size_t input_buf_size,
    char *output_buf,
    size_t output_buf_size
)
{
    ic_krn_req_t *kreqp = NULL;
    enum clnt_stat clstat;
    int wait_result;

    ASSERT(h != NULL);

    /* Build the kernel request that will be queued and waited for. */
    if ((kreqp = (ic_krn_req_t *)mdki_linux_kmalloc(sizeof(*kreqp),
                                                    KM_SLEEP)) == NULL)
    {
        clstat = RPC_SYSTEMERROR;
        goto done;
    }
    kreqp->ickr_next = NULL;
    kreqp->ickr_type = IC_PENDING; /* Where we go initially. */
    kreqp->ickr_handle = xid;
    init_waitqueue_head(&(kreqp->ickr_waitqueue));
    kreqp->ickr_res_available = FALSE;
    kreqp->ickr_res_stat = RPC_SUCCESS;
    kreqp->ickr_op = op;
    kreqp->ickr_input_bufp = input_buf;
    kreqp->ickr_input_bufsz = input_buf_size;
    kreqp->ickr_output_bufp = output_buf;
    kreqp->ickr_output_bufsz = output_buf_size;

    /* Our caller has a hold on the viewvp to which the handle we got belongs.
    ** Thus, nothing can happen to the handle until we return.  We lock the
    ** handle so we can put the request on the pending list and wakeup the ioctl
    ** waitqueue where a user space thread is (or will be) waiting for a request
    ** to handle.  Then, we unlock and wait on the waitqueue in the request
    ** until we either get a response, get a timeout, or get a shutdown (because
    ** somebody called the MVFS_CMD_RMVIEWTAG ioctl on the view to which this
    ** handle belongs).  We get the lock again and figure out why we we woke up,
    ** remove the request from any list it might be on (no list if we woke up
    ** with a response), and return the wakeup information to our caller.  At
    ** that point, we're not on any list in the handle, so we're done with the
    ** handle and nobody is waiting on the request waitqueue.
    */
    MVFS_LOCK(&(h->ic_h_lock));

    mvfs_lpc_list_add(h, kreqp);
    wake_up_interruptible(&(h->ic_ioctl_waitqueue));
    MVFS_LPC_LOG("Wakeup ioctl h=%p wq=%p n=%p p=%p.\n",
                 h, &(h->ic_ioctl_waitqueue),
                 h->ic_ioctl_waitqueue.task_list.next,
                 h->ic_ioctl_waitqueue.task_list.prev);

    MVFS_UNLOCK(&(h->ic_h_lock));

    /* Wait for a response (timeout is in .1secs, but wait wants jiffies). */
    MVFS_LPC_LOG("Wait response op=%d mh=%lld to=%ld .1secs.\n",
                 kreqp->ickr_op, kreqp->ickr_handle,
                 kreqp->ickr_waitqueue, timeout);

#if !defined(wait_event_killable_timeout)
    /* The wait_event_killable_timeout function we need didn't appear until the
    ** 4.13 kernel, but test for the interface rather than a kernel version
    ** (i.e., LINUX_VERSION_CODE < KERNEL_VERSION(4,13,0)) in case somebody
    ** backports it.  If it's not there, build it "by hand", modeled on
    ** wait_event_interruptible_timeout (which is what it looks like in 4.13).
    ** Just copy the macros with the little changes we need.  The
    ** wait_event_killable (but no timeout) interface (including TASK_KILLABLE)
    ** is defined in at least 2.6.27, which is the earliest kernel version we
    ** support.  Note, they changed the implementation details in various kernel
    ** versions, so do the changes based on kernel version...
    */
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,13,0)
#define __wait_event_killable_timeout(wq, condition, ret)               \
do {                                                                    \
        DEFINE_WAIT(__wait);                                            \
                                                                        \
        for (;;) {                                                      \
                prepare_to_wait(&wq, &__wait, TASK_KILLABLE);           \
                if (condition)                                          \
                        break;                                          \
                if (!fatal_signal_pending(current)) {                   \
                        ret = schedule_timeout_killable(ret);           \
                        if (!ret)                                       \
                                break;                                  \
                        continue;                                       \
                }                                                       \
                ret = -ERESTARTSYS;                                     \
                break;                                                  \
        }                                                               \
        finish_wait(&wq, &__wait);                                      \
} while (0)

#define wait_event_killable_timeout(wq, condition, timeout)             \
({                                                                      \
        long __ret = timeout;                                           \
        if (!(condition))                                               \
                __wait_event_killable_timeout(wq, condition, __ret);    \
        __ret;                                                          \
})
#else
/* Must be LINUX_VERSION_CODE >= KERNEL_VERSION(3,13,0)
   where they changed the macros to use ___wait_event.
*/
#define __wait_event_killable_timeout(wq, condition, timeout)		\
	___wait_event(wq, ___wait_cond_timeout(condition),		\
		      TASK_KILLABLE, 0, timeout,			\
		      __ret = schedule_timeout_killable(__ret))

#if LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0)
#define wait_event_killable_timeout(wq, condition, timeout)		\
({									\
	long __ret = timeout;						\
	if (!___wait_cond_timeout(condition))				\
		__ret = __wait_event_killable_timeout(wq,		\
						condition, timeout);	\
	__ret;								\
})
#else
/* Must be LINUX_VERSION_CODE >= KERNEL_VERSION(3,19,0) &&
           LINUX_VERSION_CODE <= KERNEL_VERSION(4,12,0)
   where they added might_sleep(), and the real interface showed up
   in 4.13
*/
#define wait_event_killable_timeout(wq, condition, timeout)		\
({									\
	long __ret = timeout;						\
	might_sleep();							\
	if (!___wait_cond_timeout(condition))				\
		__ret = __wait_event_killable_timeout(wq,		\
						condition, timeout);	\
	__ret;								\
})
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(3,19,0) */
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(3,13,0) */
#endif /* !defined(wait_event_killable_timeout) */

    /* Use wake_up (rather than wake_up_interruptible) wherever anything might
    ** change the condition.  Note, for signals, this interface only returns for
    ** fatal_signal_pending (i.e., SIGKILL), rather than any signal.
    */
    wait_result =
        wait_event_killable_timeout(
            kreqp->ickr_waitqueue,
            (kreqp->ickr_res_available || h->ic_shutting_down),
            msecs_to_jiffies(timeout * 100)
        );

    /* These are the documented reasons for returning from this wait:
    **
    ** wait_result >= 1            : condition true, jiffies remaining
    ** wait_result == 0            : condition unknown, timeout expired
    ** wait_result == -ERESTARTSYS : condition unknown, interrupted by signal
    ** wait_result == <other>      : undefined
    **
    ** We have to lock and check the condition again since, if we timed out or
    ** got a signal, it could have become true between the time we returned from
    ** the wait and when we get the lock.
    */
    MVFS_LOCK(&(h->ic_h_lock));

    if (kreqp->ickr_res_available) {
        /* The request completed, so it was removed from the sent list under the
        ** lock (so, before we could get here).  Just return the completed
        ** request's response status.  The wait_result could have been anything
        ** (in the <= 0 case, not in the >= 1 case) since the flag could change
        ** between setting the wait_result and getting the lock.  Note, the
        ** ic_shutting_down flag could also be set at this point, but we don't
        ** care since we got a good response that we can return to the caller.
        */
        clstat = kreqp->ickr_res_stat;
    } else {
        /* The request didn't complete because of a timeout or interrupt (so
        ** wait_result < 1), or we're shutting down (so wait_result >= 1), so
        ** we're sitting on the pending or sent list.  Remove the request from
        ** whichever list it's on (indicated by the ickr_type field in our
        ** request) and set the return status based on the wait return value or
        ** shutdown state.
        */
        mvfs_lpc_list_remove_req(h, kreqp);

        if (wait_result == 0) {
            clstat = RPC_TIMEDOUT;
        } else if (wait_result == -ERESTARTSYS) {
            clstat = RPC_INTR;
        } else if (h->ic_shutting_down) {
            clstat = RPC_CANTSEND;
        } else {
            clstat = RPC_SYSTEMERROR;
        }
    }
    MVFS_UNLOCK(&(h->ic_h_lock));

  done:
    MVFS_LPC_LOG("Response op=%d mh=%lld stat=%d h=%p shutting=%d.\n",
                 op, xid, clstat, h, h->ic_shutting_down);

    if (kreqp != NULL) {
        /* We don't need to cleanup anything from the init_waitqueue_head(). */
        mdki_linux_kfree(kreqp, sizeof(*kreqp));
    }
    return clstat;
}

int
mvfs_lpc_linux_ioctl(
    MVFS_LPC_MDEP_H_T h,
    void *datap,
    u_long datalen,
    CALL_DATA_T *cd,
    MVFS_CALLER_INFO *callinfo
)
{
    int err = 0;
    mvfs_lpc_usr_hdr_t uhdr;    /* Relatively small, so stack is OK. */
    mvfs_lpc_usr_msg_t *umsgp = (mvfs_lpc_usr_msg_t *)datap;

    ASSERT((h != NULL) && (datap != NULL));

    /* Verify there's at least room for a message header in the data. */
    if (datalen < sizeof(uhdr)) {
        err = EINVAL;
        mvfs_log(MFS_LOG_ERR,
                 "%s: datalen=%lu < required=%lu err=%d\n",
                 __func__, datalen, sizeof(uhdr), err);
        goto done;
    }
    /* We need to copyin the message header so we can figure out if this is a
    ** response or just a buffer ready to receive a request.  If it's a
    ** response, we only have to copyin the amount of data the header says
    ** (probably much less than the datalen).  If it is just a buffer (or after
    ** we've handled the response, at which point it is just a buffer) it has
    ** datalen bytes we can use, and we will specify how many we actually use in
    ** the header we return.
    **
    ** The message header is the same size in 32-bit and 64-bit environments, so
    ** we can just do the copyin.
    */
    if (COPYIN(datap, &uhdr, sizeof(uhdr)) != 0) {
        err = ENODATA;
        mvfs_log(MFS_LOG_ERR, "%s: COPYIN of uhdr req=%p\n", __func__, datap);
        goto done;
    }
    /* If this is a response, find the request it goes with and return the
    ** response to it (so the buffer is "empty" on return).
    */
    if (uhdr.lpc_is_response) {
        mvfs_lpc_linux_match_res_to_req(h, &uhdr, umsgp, datalen, cd);
    } else {
        MVFS_LPC_LOG("!Response ubuf=%p mh=%d."
                      "  Probably the first call to the IOCTL.\n",
                      umsgp, uhdr.krn_msg_handle);
    }
    /* At this point the passed in usr buffer is "empty" (it started empty or
    ** we've processed the response it contained) so we can give it to a waiting
    ** request on the pending list (waiting for such a request if necessary) and
    ** move the request to the sent list.  The umsgp points to the original usr
    ** space buffer (which has a hdr and the rest of it is for data).  This call
    ** will copyout the kernel RPC request to the data part of the buffer and
    ** copyout the header info (after filling it in from the request).
    */
    err = mvfs_lpc_linux_wait_for_req(h, &uhdr, umsgp, datalen, cd);
    MVFS_LPC_LOG("Request return ubuf=%p mh=%d op=%lld err=%d\n",
                 umsgp, uhdr.krn_msg_handle, uhdr.lpc_cmd, err);

  done:
    return err;
}

static void
mvfs_lpc_linux_match_res_to_req(
    MVFS_LPC_MDEP_H_T h,
    mvfs_lpc_usr_hdr_t *uhdrp,  /* Copied in header info. */
    mvfs_lpc_usr_msg_t *umsgp,  /* User space pointer. */
    u_long datalen,             /* Total length of user data. */
    CALL_DATA_T *cd
)
{
    int err = 0;
    ic_krn_req_t **link_pp;
    ic_krn_req_t *kreqp;
    enum rpc_accept_stat rpc_prot_stat;

    ASSERT((h != NULL) && (uhdrp != NULL) && uhdrp->lpc_is_response);

    /* Whatever happens (good or bad), we're done with this response.  Note, the
    ** lpc_res_rc is actually an RPC protocol defined value (hence, although
    ** names may differ, the values are the same for all RPC uses of it).  In
    ** user space it is an enum ks_accept_stat defined in <ks_rpc/rpc_msg.h>.
    ** In the kernel it is an enum rpc_accept_stat defined in
    ** <linux/sunrpc/msg_prot.h>, so we pass it between user space and the
    ** kernel as a int32_t and do this cast assignment here to get the right
    ** type.
    */
    uhdrp->lpc_is_response = FALSE;
    rpc_prot_stat = (enum rpc_accept_stat)(uhdrp->lpc_res_rc);

    /* Search the sent list for the handle we got from user space. */
    MVFS_LOCK(&(h->ic_h_lock));

    if ((kreqp = mvfs_lpc_list_remove_id(h, uhdrp->krn_msg_handle)) == NULL) {
        mvfs_log(MFS_LOG_INFO,
                 "%s: user space res=%p (handle=%d stat=%d) but no matching"
                 " kernel req (h=%p sent list=%p...discarding response)\n",
                 __func__, umsgp, uhdrp->krn_msg_handle, rpc_prot_stat,
                 h, h->ic_sent_head);
        goto unlock;
    }
    /* Even though the request is now not on any list (since we just removed it
    ** from the sent list), the caller is still waiting on it and could be woken
    ** up by its timeout or an interrupt at any time.  If so, it will get the
    ** lock and clean itself up.  So, we need to continue holding the lock
    ** because we use fields in the request that better be there when we
    ** reference them.
    **
    ** The request we found will always "receive" this response.  Set the
    ** response enum clnt_stat (defined in linux_vnode/linux_krpc.h) to indicate
    ** anything that went wrong in userspace or in our handling of the response.
    ** Note, RPC_SUCCESS is defined as a macro in linux_krpc.h (yuck), so it
    ** gets replaced by MVFS_RPC_SUCCESS here, which is an enum clnt_stat
    ** defined in that header file.  To avoid a compiler warning, we have to
    ** cast it "back" to the enum rpc_accept_stat type (which also defines
    ** RPC_SUCCESS).
    */
    if (rpc_prot_stat != (enum rpc_accept_stat)RPC_SUCCESS) {
        mvfs_log(MFS_LOG_INFO,
                 "%s: req=%p userspace failed with ks_accept_stat=%d"
                 " op=%d mh=%d\n",
                 __func__, kreqp, rpc_prot_stat,
                 kreqp->ickr_op, uhdrp->krn_msg_handle);
        switch (rpc_prot_stat) {
          case RPC_PROG_UNAVAIL: /* 1 */
            kreqp->ickr_res_stat = RPC_PROGUNAVAIL;
            break;
          case RPC_PROG_MISMATCH:/* 2 */
            kreqp->ickr_res_stat = RPC_VERSMISMATCH;
            break;
          case RPC_PROC_UNAVAIL: /* 3 */
            kreqp->ickr_res_stat = RPC_PROCUNAVAIL;
            break;
          case RPC_GARBAGE_ARGS: /* 4 */
            kreqp->ickr_res_stat = RPC_CANTDECODEARGS;
            break;
          case RPC_SYSTEM_ERR:   /* 5 */
          default:
            kreqp->ickr_res_stat = RPC_SYSTEMERROR;
            break;
        }
    }
    /* Verify that the buffer the user passed in is big enough to hold the
    ** response they specified.
    */
    else if ((datalen - sizeof(*uhdrp)) < uhdrp->lpc_res_out_sz) {
        mvfs_log(MFS_LOG_ERR,
                 "%s: malformed response"
                 " buf sz=%ld - hdr sz=%ld is < specified res sz=%ld"
                 " ...discarding response\n",
                 __func__, datalen, sizeof(*uhdrp), uhdrp->lpc_res_out_sz);
        kreqp->ickr_res_stat = RPC_CANTDECODERES;
    }
    /* Verify the response buffer we got from the requestor is big enough to
    ** hold the response we have.
    */
    else if (kreqp->ickr_output_bufsz < uhdrp->lpc_res_out_sz) {
        mvfs_log(MFS_LOG_ERR,
                 "%s: req=%p bufsz=%ld too small for response sz=%ld"
                 " ...op=%d mh=%d\n",
                 __func__, kreqp, kreqp->ickr_output_bufsz,
                 uhdrp->lpc_res_out_sz, kreqp->ickr_op, uhdrp->krn_msg_handle);
        kreqp->ickr_res_stat = RPC_CANTDECODERES;
    }
    /* Now we can safely copyin the response and check how it went. */
    else if (COPYIN(umsgp->lpc_data,
                    kreqp->ickr_output_bufp,
                    kreqp->ickr_output_bufsz) != 0)
    {
        mvfs_log(MFS_LOG_ERR,
                 "%s: COPYIN from %p to %p for %ld bytes failed"
                 "...op=%d mh=%d\n",
                 __func__, umsgp->lpc_data, kreqp->ickr_output_bufp,
                 kreqp->ickr_output_bufsz,
                 kreqp->ickr_op, uhdrp->krn_msg_handle);
        kreqp->ickr_res_stat = RPC_CANTDECODERES;
    }
    /* It's all good and the requestor's response buffer is filled in. */
    else {
        kreqp->ickr_res_stat = RPC_SUCCESS;
    }
    /* Set the flag the requester is waiting for and wake it up while still
    ** holding the lock since we need the request fields to still be there.
    */
    kreqp->ickr_res_available = TRUE;
    wake_up(&(kreqp->ickr_waitqueue));
    MVFS_LPC_LOG("Wakeup request ubuf=%p req=%p wq=%p op=%d mh=%d rc=%d.\n",
                 umsgp,
                 kreqp,
                 kreqp->ickr_waitqueue,
                 kreqp->ickr_op,
                 uhdrp->krn_msg_handle,
                 uhdrp->lpc_res_rc);

  unlock:
    MVFS_UNLOCK(&(h->ic_h_lock));
    return;
}

static int
mvfs_lpc_linux_wait_for_req(
    MVFS_LPC_MDEP_H_T h,
    mvfs_lpc_usr_hdr_t *uhdrp,
    mvfs_lpc_usr_msg_t *umsgp,
    u_long datalen,
    CALL_DATA_T *cd
)
{
    int err;
    int wait_result;
    ic_krn_req_t *kreqp = NULL;

    ASSERT((h != NULL) && (uhdrp != NULL) && !(uhdrp->lpc_is_response));

    MVFS_LOCK(&(h->ic_h_lock));

    /* Wait until there's something on the pending list (maybe it's there
    ** already) or we're shutting down.  There's no timeout so we just sit here
    ** until a request comes.  We wait interruptible so the user can end it with
    ** a signal.
    */
    while ((h->ic_pend_head == NULL) && !(h->ic_shutting_down)) {
        /* Unlock so we can wait. */
        MVFS_UNLOCK(&(h->ic_h_lock));

        MVFS_LPC_LOG("Wait ioctl h=%p wq=%p n=%p p=%p.\n",
                     h, &(h->ic_ioctl_waitqueue),
                     h->ic_ioctl_waitqueue.task_list.next,
                     h->ic_ioctl_waitqueue.task_list.prev);
        wait_result =
            wait_event_interruptible(
                h->ic_ioctl_waitqueue,
                ((h->ic_pend_head != NULL) || h->ic_shutting_down)
            );
        if (wait_result != 0) {
            /* Should be -ERESTARTSYS in this case. */
            err = vnlayer_errno_linux_to_unix(wait_result);
            goto done;
        }
        /* The condition was true, so lock and check it again. */
        MVFS_LOCK(&(h->ic_h_lock));
    }
    /* After the loop, we have the lock and something on the pending list or
    ** we're shutting down.  Note, even if we're shutting down, there could be
    ** something on the pending list, but we let that request take care of
    ** itself (i.e., remove itself from the pending list on the call side when
    ** it gets woken up as part of shutting down).
    */
    if (h->ic_shutting_down) {
        err = ESHUTDOWN;
        goto unlock;
    }
    /* Not shutting down, so we must have a request on the pending list. */
    kreqp = h->ic_pend_head;
    ASSERT(kreqp != NULL);

    /* Check that there's enough space for the pending request. */
    if ((datalen - sizeof(*uhdrp)) < kreqp->ickr_input_bufsz) {
        mvfs_log(MFS_LOG_ERR,
                 "%s: user buffer sz=%ld - hdr sz=%ld is < req sz=%ld\n",
                 __func__, datalen, sizeof(*uhdrp), kreqp->ickr_input_bufsz);

        /* Just leave the request on the pending list and return the error to
        ** user space in the hopes that it will call again with a bigger buffer.
        */
        err = E2BIG;
        goto unlock;
    }
    /* Now we just have to fill in the header and copy it out to the user space
    ** buffer and then follow it with the request copyout.  Either copyout could
    ** fail, so get everything setup, do the copyout's, and if they succeed then
    ** put the request on the sent list.  Otherwise, just leave it on the
    ** pending list and hope that the next user call has better luck.  Note,
    ** we're doing all this under the lock, so everything is consistent.
    */
    uhdrp->lpc_cmd = kreqp->ickr_op;
    uhdrp->lpc_req_sz = kreqp->ickr_input_bufsz;
    uhdrp->lpc_res_in_sz = kreqp->ickr_output_bufsz;
    uhdrp->krn_msg_handle = kreqp->ickr_handle;
    uhdrp->lpc_is_response = FALSE;

    if (COPYOUT(uhdrp, umsgp, sizeof(*uhdrp)) != 0) {
        mvfs_log(MFS_LOG_ERR, "%s: COPYOUT of uhdr=%p\n", __func__, uhdrp);
        err = ENODATA;
        goto unlock;
    }
    if (COPYOUT(kreqp->ickr_input_bufp,
                umsgp->lpc_data,
                kreqp->ickr_input_bufsz) != 0)
    {
        mvfs_log(MFS_LOG_ERR, "%s: COPYOUT of req=%p\n", __func__, kreqp);
        err = ENODATA;
        goto unlock;
    }
    /* Copyouts worked, remove from the pending list. */
    ASSERT(kreqp->ickr_type == IC_PENDING);
    mvfs_lpc_list_remove_req(h, kreqp);

    /* Add to the sent list. */
    kreqp->ickr_type = IC_SENT;
    mvfs_lpc_list_add(h, kreqp);
    err = 0;

  unlock:
    MVFS_UNLOCK(&(h->ic_h_lock));

    if (kreqp != NULL) {
        MVFS_LPC_LOG("Return ubuf=%p op=%d mh=%lld err=%d.\n",
                     umsgp, kreqp->ickr_op, kreqp->ickr_handle, err);
    } else {
        MVFS_LPC_LOG("Return ubuf=%p err=%d shutting down=%d.\n",
                     umsgp, err, h->ic_shutting_down);
    }
  done:
    return err;
}
static const char vnode_verid_mvfs_lpc_linux_c[] = "$Id:  038c27e0.c55f11e7.9b1a.00:01:84:fa:a2:18 $";
